#!/usr/local/bin/perl -w

# Computes Tunes dependencies.
#
# Currently supported:
# * provide and require or features in .scm files,
#  with at most one provided feature per file.
#  creates Makefile entries like
#	.PHONY:		feature
#	feature:	feature.scm requiredfeature1 requiredfeature2

use IO::File ;
use strict;


my $DEBUG = 0; # 1
$| = $DEBUG;

sub errprint {
    print STDERR join(" ",@_)."\n";
}
sub dbg {
    if($DEBUG) {
	errprint @_;
    }
}
sub suckinput {
    my $input = "" ;
    while (<>) {
	$input .= $_ ;
    }
    return $input ;
}
sub suckinputaslist {
    my @input = "" ;
    while (<>) {
	push @input,$_ ;
    }
    return @input ;
}


my $SCM_LOADPATH = $ENV{"SCHEME_LOADPATH"} || ".:./lib" ;
my $HLM_LOADPATH = $ENV{"HLM_LOADPATH"} ;

my @SCM_LOADPATH = (".",(split /:/, $SCM_LOADPATH)) ;
my @HLM_LOADPATH = (".","config","langC","libc","runtime") ;
my @FEATURE_LIST = ();
my %DEPENDENCIES = ();

my %DONE = () ;
my @SCM_QUEUE = () ;
my @HLM_QUEUE = () ;
my $SQP = 0 ;
my $HQP = 0 ;





dbg ((join "\n",@SCM_LOADPATH), "\n\n");
dbg ((join "\n",@HLM_LOADPATH), "\n\n");

sub lookup_path ($@) {
    my $file = shift ;
    my $path ;
    if ( -f $file ) { return $file }
    while ( $path = shift ) {
	if ( -f "$path/$file" ) { return "$path/$file" }
    }
    return
}
sub lookup_scm_path ($) { lookup_path shift,@SCM_LOADPATH ; }
sub lookup_hlm_path ($) { lookup_path shift,@HLM_LOADPATH ; }
 
sub do_queue ($\@\%) {
    my ($elem,$queue,$done) = @_ ;
    if (not defined $$done{$elem}) {
	push @$queue,$elem;
    }
}
sub do_unqueue (\@\%) {
    my ($queue,$done) = @_ ;
    my $elem = shift @$queue ;
    $$done{$elem} = 1 ;
    return $elem ;
}
sub queue_scm ($) { do_queue shift,@SCM_QUEUE,%DONE ; }
sub queue_hlm ($) { do_queue shift,@HLM_QUEUE,%DONE ; }
sub unqueue_scm ($) { do_unqueue @SCM_QUEUE,%DONE ; }
sub unqueue_hlm ($) { do_unqueue @HLM_QUEUE,%DONE ; }

sub norm_atom ($) {
    local $_ = shift ;
    s/^\'// || s/^\"(.*)\"/$1/ || s/^.*$// ;
    $_
}
sub norm_scm ($) {
    local $_ = norm_atom shift ;
    if ( /./ && ! /\.scm/) { $_ .= ".scm" }
    $_
}
sub norm_hlm ($) {
    local $_ = norm_atom shift ;
    if ( /./ && ! /\.hlm/) { $_ .= ".hlm" }
    $_
}
sub depend_scm_file ($) {
    my $file = shift;
    dbg "dsf: $file";
    my $fh = new IO::File "<$file" ;
    my @required = ();
    while(<$fh>) {
	if(/^\((load|require) *([^ ()]*)\)/i) {
	    my $scm = norm_scm $2 ;
	    if ( $scm ne "" ) {
		#dbg "detected required feature $scm in $file...";
		my $dep = lookup_scm_path $scm ;
		if ( $dep ne "" ) {
		    queue_scm $dep;
		    push @required, $dep;
		} else {
		    errprint "Could not find file $scm loaded by $file\n";
		}
	    }
	}
	if( /^\(hlm\-(load|require) *([^ ()]*)\)/i ) {
	    my $hlm = norm_hlm $2 ;
	    if ( $hlm ne "" ) {
		#dbg "detected required feature $hlm in $file...";
		my $dep = lookup_hlm_path $hlm ;
		if ( $dep ne "" ) {
		    queue_hlm $dep;
		    push @required, $dep;
		} else {
		    errprint "Could not find file $hlm required by $file\n";
		}
	    }
	}
    }
    dbg "dsfr: ",@required;
    $DEPENDENCIES{$file} = \@required ;
    $fh->close;
}
sub depend_hlm_file ($) {
    my $file = shift;
    dbg "dhf: $file";
    my $fh = new IO::File "<$file" ;
    my @required = ();
    while(<$fh>) {
	if(/^\((load|require) *([^ ()]*)\)/i) {
	    my $hlm = norm_hlm $2 ;
	    if ( $hlm ne "" ) {
		#dbg "detected required feature $hlm in $file...";
		my $dep = lookup_hlm_path $hlm ;
		if ( $dep ) {
		    queue_hlm $dep;
		    push @required, $dep;
		} else {
		    errprint "Could not find file $hlm required by $file\n";
		}
	    }
	}
    }
    dbg "dhfr: ",@required;
    $DEPENDENCIES{$file} = \@required ;
    $fh->close;
}

sub compute_graph () {
    while ( $#SCM_QUEUE + $#HLM_QUEUE > -2 ) {
	if ( $#SCM_QUEUE > -1 ) {
	    depend_scm_file &unqueue_scm ;
	}
	if ( $#HLM_QUEUE > -1 ) {
	    depend_hlm_file &unqueue_hlm ;
	}
    }
}


sub dot_depend () {
    my ($file,$deps) ;
    while (($file,$deps) = each %DEPENDENCIES) {
	print ".PHONY: $file.foo\n"
	}
    while (($file,$deps) = each %DEPENDENCIES) {
	print "$file.foo:: $file ", (join " ",map {"$_.foo"} @$deps), "\n";
    }
}

sub list_file_depend ($$) {
    dbg "xxx",@_,"\n" ;
    my $list = shift ;
    my $file = shift ;
    dbg "lfd",$list,$file,@$list ;
    if (not defined $DONE{$file}) {
	dbg "lfd2",$file ;
	$DONE{$file} = 1 ;
	my $dep = $DEPENDENCIES{$file} ;
	if (defined $dep) {
	    &list_files_depend($list,@$dep) ;
	}
	push @{$list},$file ;
    }
}

sub list_files_depend ($@) {
    dbg "lfsd0" ;
    dbg "lfsd",@_,"\n" ;
    dbg "lfsd1" ;
    my $list = shift ;
    map {&list_file_depend($list,$_)} @_ ;
}

sub print_files_depend ($@) {
    my $sep = shift ;
    my @list = () ;
    %DONE = () ;
    dbg "pfsd sep='$sep'",\@list,@list,@_,"\n" ;
    dbg "pfsd1" ;
    list_files_depend(\@list,@_);
    print (join $sep,@list) ;
}


sub merge_scm_file ($) {
    my $file = shift;
    my $text = "";
    my $fh = new IO::File "<$file" ;
    dbg "dsf: $file";
    my @required = ();
    while(<$fh>) {
	;;;;; if(/^\((load|require) *([^ ()]*)\)/i) {
	    my $scm = norm_scm $2 ;
	    if ( $scm ne "" ) {
		#dbg "detected required feature $scm in $file...";
		my $dep = lookup_scm_path $scm ;
		if ( $dep eq "" ) {
		    errprint "Could not find file $scm loaded by $file\n";
		} else {
		    if (not defined $DONE{$dep}) {
			$DONE{$dep} = 1 ;
			$text .= merge_scm_file($dep) . "\n" ;
		    }
		}
	    }
	} elsif( /^\(hlm\-(load|require) *([^ ()]*)\)/i ) {
	    my $hlm = norm_hlm $2 ;
	    if ( $hlm ne "" ) {
		#dbg "detected required feature $hlm in $file...";
		my $dep = lookup_hlm_path $hlm ;
		if ( $dep eq "" ) {
		    errprint "Could not find file $hlm required by $file\n";
		} else {
		    if (not defined $DONE{$dep}) {
			$DONE{$dep} = 1 ;
			$text .=
			    "(hlm-loaded '(" .
			    &merge_hlm_file($dep) .
			    "))\n" ;
		    }
		}
	    }
	} else {
	    $text .= $_ ;
	}
    }
    dbg "dsfr: ",@required;
    $DEPENDENCIES{$file} = \@required ;
    $fh->close;
    return $text ;
}
sub merge_hlm_file {
    my $file = shift;
    my $text  = "";
    dbg "dhf: $file";
    my $fh = new IO::File "<$file" ;
    my @required = ();
    while(<$fh>) {
	if(/^\((load|require) *([^ ()]*)\)/i) {
	    my $hlm = norm_hlm $2 ;
	    if ( $hlm ne "" ) {
		#dbg "detected required feature $hlm in $file...";
		my $dep = lookup_hlm_path $hlm ;
		if ( $dep eq "" ) {
		    errprint "Could not find file $hlm required by $file\n";
		} else {
		    if (not defined $DONE{$dep}) {
			$DONE{$dep} = 1 ;
			$text .=
			    "(loaded '(" .
			    &merge_hlm_file($dep) .
			    "))\n" ;
		    }
		}
	    }
	}
    }
    dbg "dhfr: ",@required;
    $DEPENDENCIES{$file} = \@required ;
    $fh->close;
    return $text ;
}

sub merge (@) {
    print (join "\n",(map {merge_scm_file $_} @_));
}


my $fun = shift @ARGV ;
$_ = $fun ;
;;;; if ( $fun eq "depend" ) {
    @SCM_QUEUE = @ARGV ;
    &compute_graph ;
    &dot_depend() ;
} elsif ( $fun eq "merge" ) {
    &merge(@ARGV) ;
} elsif ( $fun eq "deps" ) {
    @SCM_QUEUE = @ARGV ;
    &compute_graph ;
    &print_files_depend(" ",@ARGV) ;
} elsif ( $fun eq "depln" ) {
    @SCM_QUEUE = @ARGV ;
    &compute_graph ;
    &print_files_depend("\n",@ARGV) ;
    print "\n" ;
} else {
    print STDERR "Unknown command $fun\n" ;
    exit 10 ;
}
